#ifndef COMMON_CK
#define COMMON_CK

#include "CalcIntgrt.h"

using namespace myLib;

namespace cmmn {


	extern double * smpls_rho, ** smpls_Cx, ** smpls_Cy;
	extern 	int n_smpls;



	//
	//								Quick Sort
	//
	//      S. Kawai (1988).  Methods of programming (in Japanese). Tokyo: Iwanami-shoten.
	//		S. Ishihata (1989). Algorithms and Data Structure (in Japanese). Tokyo:Iwanami-shoten
	//
	void qSort(double * d, int Lb, int Ub);

	double CalcCor(int * iX, int * iY, int n);
	
	double phi(double x);

	double Cum_NormalKC(double z);

	inline double sqr(double x){
		return x * x;
	}

	/*
	              |           |
	       Pnp    |    Pcp    |    Ppp
		          |           |
	--------------------------------------------
	              |           |
	       Pnc    |    Pcc    |    Ppc
	              |           |
	--------------------------------------------
	              |           |
	       Pnn    |    Pcn    |    Ppn
	              |           |
	*/

	class Funcfpp {
	public:
		double rho;
		double b;

		Funcfpp(){}
		Funcfpp(double vrho, double vb) : rho(vrho), b(vb) {}

		double Fx(double x){
			double v = 1.0 - Cum_NormalKC((b - rho * x) / Math::Sqrt(1.0 - sqr(rho)));
			return v;
		}

		double Fxfx(double x){
			double v = Fx(x) * phi(x);
			return v;
		}

		double operator()(double t){
			double x = -Math::Log(-t);
			double v = Fxfx(x) * (-1.0 / t);
			return v;
		}
	};

	double CalcPpp(double vrho, double va, double vb);

	class Funcfpc {
	public:
		double rho;
		double a;

		Funcfpc(){}
		Funcfpc(double vrho, double va) : rho(vrho), a(va) {}

		double Gy(double y){
			double v = 1.0 - Cum_NormalKC((a - rho * y) / Math::Sqrt(1.0 - sqr(rho)));
			return v;
		}

		double operator()(double y){
			double v = Gy(y) * phi(y);
			return v;
		}
	};

	double CalcPpc(double vrho, double va, double vb1, double vb2);

	class Funcfpn {
	public:
		double rho;
		double b;

		Funcfpn(){}
		Funcfpn(double vrho, double vb) : rho(vrho), b(vb) {}

		double Fx(double x){
			double v = Cum_NormalKC((b - rho * x) / Math::Sqrt(1.0 - sqr(rho)));
			return v;
		}

		double Fxfx(double x){
			double v = Fx(x) * phi(x);
			return v;
		}

		double operator()(double t){
			double x = -Math::Log(-t);
			double v = Fxfx(x) * (-1.0 / t);
			return v;
		}
	};

	double CalcPpn(double vrho, double va, double vb);

	class Funcfcp {
	public:
		double rho;
		double b;

		Funcfcp(){}
		Funcfcp(double vrho, double vb) : rho(vrho), b(vb) {}

		double Fx(double x){
			double v = 1.0 - Cum_NormalKC((b - rho * x) / Math::Sqrt(1.0 - sqr(rho)));
			return v;
		}

		double operator()(double x){
			double v = Fx(x) * phi(x);
			return v;
		}
	};

	double CalcPcp(double vrho, double va1, double va2, double vb);

	class Funcfcc {
	public:
		double rho;
		double b1, b2;

		Funcfcc(){}
		Funcfcc(double vrho, double vb1, double vb2) : rho(vrho), b1(vb1), b2(vb2) {}

		double Fx(double x){
			double v = Cum_NormalKC((b2 - rho * x) / Math::Sqrt(1.0 - sqr(rho)))
						- Cum_NormalKC((b1 - rho * x) / Math::Sqrt(1.0 - sqr(rho)));
			return v;
		}

		double operator()(double x){
			double v = Fx(x) * phi(x);
			return v;
		}
	};

	double CalcPcc(double vrho, double va1, double va2, double vb1, double vb2);

	class Funcfcn {
	public:
		double rho;
		double b;

		Funcfcn(){}
		Funcfcn(double vrho, double vb) : rho(vrho), b(vb) {}

		double Fx(double x){
			double v = Cum_NormalKC((b - rho * x) / Math::Sqrt(1.0 - sqr(rho)));
			return v;
		}

		double operator()(double x){
			double v = Fx(x) * phi(x);
			return v;
		}
	};

	double CalcPcn(double vrho, double va1, double va2, double vb);

	class Funcfnp {
	public:
		double rho;
		double b;

		Funcfnp(){}
		Funcfnp(double vrho, double vb) : rho(vrho), b(vb) {}

		double Fx(double x){
			double v = 1.0 - Cum_NormalKC((b - rho * x) / Math::Sqrt(1.0 - sqr(rho)));
			return v;
		}

		double Fxfx(double x){
			double v = Fx(x) * phi(x);
			return v;
		}

		double operator()(double t){
			double x = Math::Log(t);
			double v = Fxfx(x) / t;
			return v;
		}
	};

	double CalcPnp(double vrho, double va, double vb);

	class Funcfnc {
	public:
		double rho;
		double a;

		Funcfnc(){}
		Funcfnc(double vrho, double va) : rho(vrho), a(va) {}

		double Gy(double y){
			double v = Cum_NormalKC((a - rho * y) / Math::Sqrt(1.0 - sqr(rho)));
			return v;
		}

		double operator()(double y){
			double v = Gy(y) * phi(y);
			return v;
		}
	};

	double CalcPnc(double vrho, double va, double vb1, double vb2);

	class Funcfnn {
	public:
		double rho;
		double b;

		Funcfnn(){}
		Funcfnn(double vrho, double vb) : rho(vrho), b(vb) {}

		double Fx(double x){
			double v = Cum_NormalKC((b - rho * x) / Math::Sqrt(1.0 - sqr(rho)));
			return v;
		}

		double Fxfx(double x){
			double v = Fx(x) * phi(x);
			return v;
		}

		double operator()(double t){
			double x = Math::Log(t);
			double v = Fxfx(x) / t;
			return v;
		}
	};

	double CalcPnn(double vrho, double va, double vb);

	class CalcPxy {
	public:
		double rho;
		int Kx, Ky;
		double * Cx, * Cy;
		double ** Pxy;

		CalcPxy(){}
		CalcPxy(double vrho, int vKx, int vKy, double * vCx, double * vCy) :
			rho(vrho), Kx(vKx), Ky(vKy) {
			Cx = new double[Kx];
			for (int i = 0; i < Kx - 1; i++)
				Cx[i] = vCx[i];
			Cy = new double[Ky];
			for (int j = 0; j < Ky - 1; j++)
				Cy[j] = vCy[j];
			Pxy = new double*[Kx];
			for (int i = 0; i < Kx; i++)
				Pxy[i] = new double[Ky];

			calculate();
		}
		virtual ~CalcPxy(){
			if (Cx != nullptr) delete[] Cx;
			if (Cy != nullptr) delete[] Cy;
			if (Pxy != nullptr){
				for (int i = 0; i < Kx; i++){
					delete[] Pxy[i];
				}
				delete[] Pxy;
			}
		}

		void calculate();
	};

	class CalcPxyCx {
	public:
		int Ck;
		double rho;
		int Kx, Ky;
		double * Cx, *Cy;
		double ** Pxy;

		CalcPxyCx(){}
		CalcPxyCx(int vCk, double vrho, int vKx, int vKy, double * vCx, double * vCy) :
			Ck(vCk), rho(vrho), Kx(vKx), Ky(vKy) {
			Cx = new double[Kx];
			for (int i = 0; i < Kx - 1; i++)
				Cx[i] = vCx[i];
			Cy = new double[Ky];
			for (int j = 0; j < Ky - 1; j++)
				Cy[j] = vCy[j];
			Pxy = new double*[Kx];
			for (int i = 0; i < Kx; i++)
				Pxy[i] = new double[Ky];

			calculate();
		}
		virtual ~CalcPxyCx(){
			if (Cx != nullptr) delete[] Cx;
			if (Cy != nullptr) delete[] Cy;
			if (Pxy != nullptr){
				for (int i = 0; i < Kx; i++){
					delete[] Pxy[i];
				}
				delete[] Pxy;
			}
		}

		void calculate();
	};


	class CalcPxyCy {
	public:
		int Ch;
		double rho;
		int Kx, Ky;
		double * Cx, *Cy;
		double ** Pxy;

		CalcPxyCy(){}
		CalcPxyCy(int vCh, double vrho, int vKx, int vKy, double * vCx, double * vCy) :
			Ch(vCh), rho(vrho), Kx(vKx), Ky(vKy) {
			Cx = new double[Kx];
			for (int i = 0; i < Kx - 1; i++)
				Cx[i] = vCx[i];
			Cy = new double[Ky];
			for (int j = 0; j < Ky - 1; j++)
				Cy[j] = vCy[j];
			Pxy = new double*[Kx];
			for (int i = 0; i < Kx; i++)
				Pxy[i] = new double[Ky];

			calculate();
		}
		virtual ~CalcPxyCy(){
			if (Cx != nullptr) delete[] Cx;
			if (Cy != nullptr) delete[] Cy;
			if (Pxy != nullptr){
				for (int i = 0; i < Kx; i++){
					delete[] Pxy[i];
				}
				delete[] Pxy;
			}
		}

		void calculate();
	};



}

#endif
