#include "mcmc.h"
#include "DispMessage.h"
#include "common.h" 

using namespace EstPolychoric;
using namespace cmmn;

namespace MyMCMC {

	inline double sqr(double x){
		return x * x;
	}

	double MyLog(double x){
		double v;
		if (x > 1.0e-308){
			v = Math::Log(x);
		}
		else{
			v = Math::Log(1.0e-308);
		}
		return v;
	}

	double MyExp(double x){
		double v;
		if (x < 709.0){
			v = Math::Exp(x);
		}
		else{
			v = Math::Exp(709.0);
		}
		return v;
	}


	void MCMCPolyChoric::ClearMe(){
		if(rn != nullptr) delete rn;
		if (Nkh != nullptr){
			for (int k = 0; k < Kx; k++)
				delete[] Nkh[k];
			delete[] Nkh;
		}
		if (X != nullptr) delete[] X;
		if (Y != nullptr) delete[] Y;
		if (rho != nullptr) delete[] rho;
		if (Cx != nullptr){
			for (int i = 0; i <= NSimu; i++)
				delete[] Cx[i];
			delete[] Cx;
		}
		if (Cy != nullptr){
			for (int i = 0; i <= NSimu; i++)
				delete[] Cy[i];
			delete[] Cy;
		}
		if (cntCx != nullptr)
			delete[] cntCx;
		if (cntCy != nullptr)
			delete[] cntCy;
		if (sgmCx != nullptr)
			delete[] sgmCx;
		if (sgmCy != nullptr)
			delete[] sgmCy;
	}

	void MCMCPolyChoric::Init(String ^ vhstr, int vnjump, int vNSimu, int vKx, int vKy, int vN, int * vX, int * vY,
		double init_rho, double * init_Cx, double * init_Cy,
		double vsgm_rho, double * vsgmCx, double * vsgmCy){
		hstr = vhstr;
		rn = new RNP2to191();
		if (vnjump > 0)
			for (int i = 0; i < vnjump; i++)
				rn->Jump2toj(100);
		NSimu = vNSimu;
		Kx = vKx;
		Ky = vKy;
		N = vN;
		Nkh = new int*[Kx];
		for (int k = 0; k < Kx; k++){
			Nkh[k] = new int[Ky];
			for (int h = 0; h < Ky; h++)
				Nkh[k][h] = 0;
		}

		X = new int[N];
		Y = new int[N];
		for (int i = 0; i < N; i++){
			X[i] = vX[i];
			Y[i] = vY[i];
			Nkh[X[i] - 1][Y[i] - 1]++;
		}

		rho = new double[NSimu + 1];
		rho[0] = init_rho;
		Cx = new double*[NSimu + 1];
		Cy = new double*[NSimu + 1];
		for (int i = 0; i <= NSimu; i++){
			Cx[i] = new double[Kx - 1];
			Cy[i] = new double[Ky - 1];
		}
		for (int k = 0; k < Kx - 1; k++)
			Cx[0][k] = init_Cx[k];
		for (int h = 0; h < Ky - 1; h++)
			Cy[0][h] = init_Cy[h];
		cnt_rho = 0;
		cntCx = new int[Kx - 1];
		for (int k = 0; k < Kx - 1; k++)
			cntCx[k] = 0;
		cntCy = new int[Ky - 1];
		for (int h = 0; h < Ky - 1; h++)
			cntCy[h] = 0;
		sgm_rho = vsgm_rho;
		sgmCx = new double[Kx - 1];
		for (int k = 0; k < Kx - 1; k++)
			sgmCx[k] = vsgmCx[k];
		sgmCy = new double[Ky - 1];
		for (int h = 0; h < Ky - 1; h++)
			sgmCy[h] = vsgmCy[h];
	}

	double MCMCPolyChoric::LProbAll(double vrho, double * vCx, double * vCy){
		CalcPxy Pxy(vrho, Kx, Ky, vCx, vCy);
		double v = 0.0;
		for (int k = 0; k < Kx; k++)
			for (int h = 0; h < Ky; h++)
				v += Nkh[k][h] * MyLog(Pxy.Pxy[k][h]);
		return v;
	}

	double MCMCPolyChoric::LProbCx(int vCk, double * vCx, int imcmc){
		CalcPxyCx Pxy(vCk, rho[imcmc + 1], Kx, Ky, vCx, Cy[imcmc]);
		double v = 0.0;
//		for (int k = 0; k < Kx; k++)
		for (int k = vCk; k <= vCk + 1; k++)
			for (int h = 0; h < Ky; h++)
				v += Nkh[k][h] * MyLog(Pxy.Pxy[k][h]);
		return v;
	}

	double MCMCPolyChoric::LProbCy(int vCh, double * vCy, int imcmc){
		CalcPxyCy Pxy(vCh, rho[imcmc + 1], Kx, Ky, Cx[imcmc + 1], vCy);
		double v = 0.0;
		for (int k = 0; k < Kx; k++)
			for (int h = vCh; h <= vCh + 1; h++)
				v += Nkh[k][h] * MyLog(Pxy.Pxy[k][h]);
		return v;
	}

	void MCMCPolyChoric::do_mcmc(){
		DispMessage dmfm;
		dmfm.Show();
		dmfm.SetMsg(hstr);
		dmfm.ShowMessage("do_mcmc started.");

		for (int istep = 0; istep < NSimu; istep++){
			if ((istep % 100) == 0){
				dmfm.ShowMessage("\n" + istep.ToString() + "/" + NSimu.ToString());
				Application::DoEvents();
			}

			double y = rn->normalMS(rho[istep], sgm_rho);
			double a;
			if ((y <= -1.0) || (y >= 1.0)){
				a = -1.0;
			}
			else{
				double num = LProbAll(y, Cx[istep], Cy[istep]);
				double den = LProbAll(rho[istep], Cx[istep], Cy[istep]);
				a = MyExp(num - den);
			}
			if (rn->uni() < a){
				rho[istep + 1] = y;
				cnt_rho++;
			}
			else{
				rho[istep + 1] = rho[istep];
			}

			double * Cx0 = new double[Kx - 1], *Cx1 = new double[Kx - 1];
			for (int i = 0; i < Kx - 1; i++)
				Cx0[i] = Cx1[i] = Cx[istep][i];
			for (int i = 0; i < Kx - 1; i++){
				double y = rn->normalMS(Cx0[i], sgmCx[i]);
				double a = 0.0;
				if (i == 0){
					if (Kx > 2)
						if (y >= Cx0[1])
							a = -1.0;
				}
				else if (i == (Kx - 2)){
					if (Kx > 2)
						if (y <= Cx0[i - 1])
							a = -1.0;
				}
				else{
					if ((y <= Cx0[i - 1]) || y >= Cx0[i + 1])
						a = -1.0;
				}
				if (a > -0.5){
					Cx1[i] = y;
					double num = LProbCx(i, Cx1, istep);
					double den = LProbCx(i, Cx0, istep);
					a = MyExp(num - den);
				}
				if (rn->uni() < a){
					Cx0[i] = y;
					cntCx[i]++;
				}
				else{
					Cx1[i] = Cx0[i];
				}
			}
			for (int i = 0; i < Kx - 1; i++)
				Cx[istep + 1][i] = Cx1[i];
			delete[] Cx0;
			delete[] Cx1;


			double * Cy0 = new double[Ky - 1], *Cy1 = new double[Ky - 1];
			for (int i = 0; i < Ky - 1; i++)
				Cy0[i] = Cy1[i] = Cy[istep][i];
			for (int i = 0; i < Ky - 1; i++){
				double y = rn->normalMS(Cy0[i], sgmCy[i]);
				double a = 0.0;
				if (i == 0){
					if (Ky > 2)
						if (y >= Cy0[1])
							a = -1.0;
				}
				else if (i == (Ky - 2)){
					if (Ky > 2)
						if (y <= Cy0[i - 1])
							a = -1.0;
				}
				else{
					if ((y <= Cy0[i - 1]) || y >= Cy0[i + 1])
						a = -1.0;
				}
				if (a > -0.5){
					Cy1[i] = y;
					double num = LProbCy(i, Cy1, istep);
					double den = LProbCy(i, Cy0, istep);
					a = MyExp(num - den);
				}
				if (rn->uni() < a){
					Cy0[i] = y;
					cntCy[i]++;
				}
				else{
					Cy1[i] = Cy0[i];
				}
			}
			for (int i = 0; i < Ky - 1; i++)
				Cy[istep + 1][i] = Cy1[i];
			delete[] Cy0;
			delete[] Cy1;
		}

		dmfm.Close();
	}

}
